"""
dk_Met_IO is a package to read or retrieve meteorological data
from various data sources.
"""

__author__ = "The R & D Center for Weather Forecasting Technology in NMC, CMA"
__version__ = '0.1.10.1'
